import re
from functools import total_ordering
from typing import Any, Dict, List, Optional, Tuple, Union, ClassVar

from jinja2 import Environment, FileSystemLoader, Template
from jinja2.exceptions import TemplateNotFound
from pydantic import BaseModel, ConfigDict

from serial_bus.config import get_config
from serial_bus.datastore import (
    SerialBusSortedSet,
    ModelsGlobalStore,
    get_global_data_store,
)
from serial_bus.exceptions import (
    SerialBusTypeError,
    ModelInitializationError,
    RenderableTemplateError,
)
from serial_bus.utils import convert_dict_to_hashabledict

GLOBAL_CONFIGS = get_config()


@total_ordering
class SerialBusBaseModel(BaseModel):
    """A pydantic BaseModel class that provides the basic
    functionality for all SerialBus Models classes.

    Any SerialBusBaseModel child class SHOULD be instantiated from
    .create_from_loaded_data() or .create() class methods provided.
    Instantiation by using __init__() directly is discouraged.

    This class implements functools.total_ordering to allow its instances
    to be sorted accordingly inside a datastore.SerialBusSortedSet.

    Because datastore.SerialBusSortedSet is a python Set, all
    SerialBus models MUST be hashable. This has a twofold consequence:
        - Model classes must implement the __hash__() method;
        - all attributes of any SerialBus model class must be hashable;
    """

    # this class attribute is used as a registry for all subclasses
    # of this one. This is useful to allow transparent discovery of
    # such classes, which in turn allows for easy mapping of directives
    # to model classes.
    _subclasses: ClassVar = []

    # class attribute to store a reference to the global data structure
    # where all model instances are stored.
    _data_store: ClassVar = get_global_data_store()

    # this is required to allow fields of non-native python types
    model_config = ConfigDict(arbitrary_types_allowed=True)

    # represents the keyword in the serialized data used to
    # trigger the instantiation of this model class
    _directive: str = None

    # a tuple of strings with each representing the name of an attribute of
    # the model class. This tuple will be used to create a unique identifier
    # that will ID this model object in the store.This will also be used as
    # a sort key in the data store. Child classes MUST override this attribute.
    _key: Tuple[str]

    # the global data store doesn't allow for duplication, as it
    # implements a set. However, SerialBus doesn't necessarily care if a
    # duplication happens.This attribute indicates if an exception should
    # be raised in case an attempted duplication is detected.
    _err_on_duplicate: bool = False

    def __new__(cls, *args, **kwargs):
        """Overrides the __new__ method to prevent direct instantiation of this class."""
        if cls == SerialBusBaseModel:
            raise ModelInitializationError(
                "Cannot instantiate SerialBusBaseModel directly."
            )
        return super().__new__(cls)

    def __init_subclass__(cls, **kwargs):
        """Registers all subclasses of this class in the _subclasses attribute."""
        super().__init_subclass__(**kwargs)
        cls._subclasses.append(cls)

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))

    def __lt__(self, other):
        return self.key <= other.key

    def __eq__(self, other):
        return str(self) == str(other)
    
    @classmethod
    def _normalize_for_validations(cls, dict_args: Dict[str, Any]) -> Dict[str, Any]:
        """
        Converts all values of dict_args to tuples when:
            1 - the value is a list object; AND
            2 - if the key associated with the value matches a filed in the model class; AND
            3 - the type associated with the field is of base type 'tuple'.
        
        This is done in the spirit of best-effort to ensure the hash-ability of the model
        instance in a seamless way. Although this helps, ideally, users should
        guarantee this by themselves by passing tuples instead of lists when the field of 
        the model class is of type tuple.

        TODO: expand on this to allow greater flexibility. For now, it's just a
        placeholder to ensure the dict_args is hashable. In the future, this
        class should allow a custom type of 'hashable sequence' to be defined
        as a means to represent ManyToOne relationships and then automatically
        convert collection types (like lists or sets) to this custom type.

        Returns:
            Dict[str, Any]: the converted dict.
        """
        converted_dict = {}

        for key, value in dict_args.items():
            if isinstance(value, list) and key in cls.__annotations__ and getattr(cls.__annotations__[key], '__origin__', None) is tuple:
                value = tuple(value)
                    
            converted_dict[key] = value

        return converted_dict
    
    @property
    def directive(self) -> str:
        """
        Returns the value of the `_directive` attribute. This is used to
        identify the model class when parsing a file of supported format.
        """
        return self._directive

    @property
    def key(self) -> Tuple:
        """
        Returns a tuple containing the values associate with the fields (attributes)
        named in the `_key` attribute.
        """
        return tuple([getattr(self, attr) for attr in self._key])

    @classmethod
    def create_from_loaded_data(
        cls, data: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> None:
        """Factory class method used to instantiate a model with data
        previously loaded and parsed from a file.

        Args:
            data (Union[Dict[str, Any], List[Dict[str, Any]]]): Data loaded
            from a file of supported format and parsed into a Dictionary or List.

        Raises:
            SerialBusTypeError: If `data` is not a dict nor a list.
        """
        if not isinstance(data, dict) and not isinstance(data, list):
            raise SerialBusTypeError(
                f"Data passed to {cls.__name__} must be of type 'dict' or 'list', but was {type(data)}"
            )

        if isinstance(data, list):
            for _dict in data:
                cls.create(dict_args=_dict)
            return

        cls.create(dict_args=data)

    @classmethod
    def create(
        cls, dict_args: Dict[Any, Any], *args, **kwargs
    ) -> "SerialBusBaseModel":
        """
        Factory class method used to instantiate a model with data
        passed as a dictionary.

        Args:
            dict_args (Dict[Any, Any]): a dictionary containing the data to be used
            to instantiate the model.

        Returns:
            SerialBusBaseModel: the instance of the model created.
        """
        dict_args = convert_dict_to_hashabledict(dict_args)
        dict_args = cls._normalize_for_validations(dict_args)
        new_obj_model = cls.model_validate(dict_args, strict=True, *args, **kwargs)
        new_obj_model.ds().save(new_obj_model)

        return new_obj_model

    @classmethod
    def ds(cls) -> ModelsGlobalStore:
        """Returns the global data store where all models are stored."""
        return cls._data_store

    @classmethod
    def filter(cls, search_params: Dict[Any, Any]) -> SerialBusSortedSet:
        """
        Returns a SerialBusSortedSet containing all instances of this
        class that match the search_params.

        Args:
            search_params (Dict[Any, Any]): a dictionary with the key being the
            attribute of the model and the value being the value to be searched for.

        Returns:
            SerialBusSortedSet: the SerialBusSortedSet containing the records
            that match the search_params.
        """
        return cls.ds().filter(cls, search_params)

    @classmethod
    def get(cls, search_params: Dict[Any, Any]) -> "SerialBusBaseModel":
        """
        Retrieves a single instance of this class that matches the provided search parameters.

        Args:
            search_params (Dict[Any, Any]): A dictionary containing the search parameters. The
            keys should correspond to the attribute names of the model, and the values should
            correspond to the expected values of these attributes.

        Returns:
            SerialBusBaseModel: The model instance that matches the search parameters.

        Raises:
            ModelDoesNotExist: If no instance of this class matches the search parameters.
            ModelAlreadyExists: If more than one instance of this class matches the search parameters.
        """
        return cls.ds().get(cls, search_params)

    @classmethod
    def get_all(cls) -> SerialBusSortedSet:
        """Returns a SerialBusSortedSet containing all instances of this class."""
        return cls.ds().get_all_by_class(cls)

    @classmethod
    def get_subclasses(cls):
        """Returns a list of all subclasses of this class."""
        return cls._subclasses

    @classmethod
    def get_subclasses_with_directive(cls):
        """
        Returns a list of all subclasses that have a `_directive` attribute set.
        The .get_default() method is used here since sub_cls._directive is
        inherently a pydantic ModelPrivateAttr.
        """
        return [
            sub_cls
            for sub_cls in cls.get_subclasses()
            if sub_cls._directive.get_default()
        ]


class SerialBusRenderableModel(SerialBusBaseModel):
    """A Renderable model allows for its instances to render a string
    according to a Jinja2 template. This enables various automation
    use-cases.

    The expected template file name is derived in either of two ways:
        - a statically defined `_template_name` attribute in the child class.
        - the class name split on the capital letters, removing
        the word 'Model' (if present), and then joined by underscores ('_'):
            - MyClass -> my_class
            - OtherClassModel -> other_class
            - ModelYetAnotherClass -> yet_another_class

    NOTE: Support for additional templating languages is under consideration.
    """

    def __new__(cls, *args, **kwargs):
        """Overrides the __new__ method to prevent direct instantiation of this class."""
        if cls == SerialBusRenderableModel:
            raise ModelInitializationError(
                "Cannot instantiate SerialBusRenderableModel directly."
            )
        return super().__new__(cls)

    @classmethod
    def create(
        cls, dict_args: Dict[Any, Any], *args, **kwargs
    ) -> "SerialBusRenderableModel":
        """Overrides the create method to guarantee the _template_name attribute"""
        cls._set_template(**dict_args)
        return super().create(dict_args, *args, **kwargs)

    @classmethod
    def _set_template(cls, **kwargs):
        """
        Guarantees the instance will have a _template_name attribute already set.
        If the child class doesn't define it already, one will be created.
        """
        try:
            getattr(cls, "_template_name")
        except AttributeError:
            informed_template_name = kwargs.get("template_name")

            if informed_template_name:
                cls._template_name = informed_template_name

            else:
                split_cls_name = re.findall("[A-Z][^A-Z]*", cls.__name__)

                if "Model" in split_cls_name:
                    split_cls_name.remove("Model")

                cls._template_name = "_".join(split_cls_name).lower()

    @property
    def template_name(self) -> str:
        """A property to access the _template_name attribute."""
        return self._template_name

    def get_rendered_str(self, extra_vars_dict: Optional[Dict[str, Any]] = None) -> str:
        """
        Renders this RenderableSerialBusModel into a string according to the Jinja2
        template indicated by _template_name attribute.

        A dictionary representation of this instance will be passed to the render method.
        However, if the template requires additional variables that wouldn't match any of
        the attributes of this instance, the caller can pass them through `extra_vars_dict`
        argument.

        Args:
            extra_vars_dict (Optional[Dict[str, Any]], optional): A dictionary containing
            extra vars that the template may require. Defaults to None.

        Returns:
            str: the rendered string produced by the .render() method of the Template
            object.
        """
        if extra_vars_dict:
            _dict_to_render = dict(self)
            _dict_to_render.update(extra_vars_dict)
            return self._get_template().render(_dict_to_render)

        return self._get_template().render(dict(self))

    def _get_template(self) -> Template:
        """
        Retrieves the template file to be used to render this RenderableSerialBusModel.
        Currently, only Jinja2 templates are supported. Hence, the file is expected to
        have a '.j2' extension.

        Raises:
            RenderableTemplateError: if a template file with expected name does not exist.

        Returns:
            Template: The template file as a jinja2.Template object.
        """
        env = Environment(loader=FileSystemLoader(GLOBAL_CONFIGS.templates_dir))
        try:
            return env.get_template(f"{self.template_name}.j2")
        except TemplateNotFound as err:
            raise RenderableTemplateError(err.message)



import importlib
import inspect
import os
from types import ModuleType
from typing import List, Dict, Type

from serial_bus.exceptions import SerialBusImportError

# Below, we define the default values for the environment variables that
# will be used to configure the package. If the user wants to override these
# values, they can do so by setting the environment variables with the same
# names as the ones defined below.

# Directory where templates are stored by default
TEMPLATES_DIR = "templates"

# The python dotted path to the class used for creating hashable dictionaries
HASHABLE_DICT_CLS = "serial_bus.custom_collections.HashableDict"

# The python dotted path to the module the provides loader functions
LOADERS_MODULE = "serial_bus.loaders"

# The python dotted path to the module that provides dumper functions
DUMPERS_MODULE = "serial_bus.dumpers"

# The python dotted paths to all modules that contain subclasses of SerialBusBaseModel
# as an environment variable, it's a comma-separated string of dotted paths.
# Example: "my_module1.models,my_module2.models"
MODELS_MODULES = "serial_bus.models"


class SerialBusConfig:
    """
    This class is used to hold the configuration of the package, which is
    determined by the environment variables.
    """

    # this attr is initialized as an empty dictionary and will be populated
    # whenever the self.directive_to_model_mapping property is called by
    # utils.create_all_models_from_dict function.
    _directive_to_model_mapping: Dict[str, Type] = {}

    @property
    def templates_dir(self) -> str:
        return os.environ.get("TEMPLATES_DIR", TEMPLATES_DIR)

    @property
    def hashable_dict_cls(self) -> str:
        return os.environ.get("HASHABLE_DICT_CLS", HASHABLE_DICT_CLS)

    @property
    def loaders_module_path(self) -> str:
        return os.environ.get("LOADERS_MODULE", LOADERS_MODULE)

    @property
    def loaders_module(self) -> ModuleType:
        try:
            loaders_module = importlib.import_module(self.loaders_module_path)
        except ImportError as err:
            raise SerialBusImportError(err)

        return loaders_module

    @property
    def dumpers_module_path(self) -> str:
        return os.environ.get("DUMPERS_MODULE", DUMPERS_MODULE)

    @property
    def dumpers_module(self) -> ModuleType:
        try:
            dumpers_module = importlib.import_module(self.dumpers_module_path)
        except ImportError as err:
            raise SerialBusImportError(err)

        return dumpers_module

    @property
    def supported_formats(self) -> List[str]:
        """
        For simplicity, we only take loaders as the basis to determine the supported formats.

        In this package, it's guaranteed to always exist a matching dumper for each loader.
        If you are using a custom loader module, you should also strive for the same.
        """
        supported_formats = [
            name.partition("_")[0]
            for name, obj in inspect.getmembers(self.loaders_module)
            if inspect.isfunction(obj) and "_loader" in name
        ]

        return supported_formats

    @property
    def models_modules(self) -> List[str]:
        return os.environ.get("MODELS_MODULES", MODELS_MODULES).split(",")

    @property
    def directive_to_model_mapping(self) -> Dict[str, type]:
        """
        Returns a dictionary with the directive names as keys and the dotted paths
        to the corresponding SerialBusBaseModel subclass as values.

        Notice that this property returns self._directive_to_model_mapping, which will
        be set once and always return that same value. 

        It's only if self._directive_to_model_mapping is empty that this method will
        try to populate it: 
            - 1st: by importing all modules in self.models_modules; and
            - 2nd: by iterating over all subclasses of SerialBusBaseModel that 
              have the '_directive' field set.
        """
        if self._directive_to_model_mapping:
            return self._directive_to_model_mapping

        self._import_all_models_modules()

        # import here to avoid circular imports error
        from serial_bus.models import SerialBusBaseModel

        for cls in SerialBusBaseModel.get_subclasses_with_directive():
            self._directive_to_model_mapping[cls._directive.get_default()] = cls

        return self._directive_to_model_mapping

    def _import_all_models_modules(self) -> None:
        """
        By importing all models returned by self.models_modules, this method ensures
        subclasses of SerialBusBaseModel are registered for discovery.

        Notice this method should be treated as private. Refrain from calling it
        directly, as it's called automatically by self.directive_to_model_mapping.

        Raises:
            SerialBusImportError: If any of the modules in self.models_modules
            cannot be imported.
        """
        for module in self.models_modules:
            try:
                importlib.import_module(module)
            except ImportError as err:
                raise SerialBusImportError(err)


GLOBAL_CONFIGS = None


def get_config():
    global GLOBAL_CONFIGS

    if GLOBAL_CONFIGS is None:
        GLOBAL_CONFIGS = SerialBusConfig()

    return GLOBAL_CONFIGS



"""
This module contains functions for creating serialized data in all supported formats.

Each function takes a Python dictionary, serializes the data, and writes it 
to a TextStream object. Optionally, a file can be created with the contents
of that TextStream if the dumper function is provided with a file path, which
can be either a Path or a str object.

If your project uses file formats other than the ones supported here, you can
add support for them by having your own python dumpers.py module and creating an
environment variable `DUMPERS_MODULE` that points to its dotted path.

Contrary to the loaders module, there's no mechanism for automatically detecting
which dumper function to use. A caller must be explicit about the dumper function
for the desired serialization format.
"""

import configparser
import io
import json
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Union

import toml
import yaml

from serial_bus.exceptions import SerialBusDumperError


def stream_dumper(dump_func: Callable[..., None]):
    def standard_dumper_func(
        data: Dict[Any, Any],
        file_path: Optional[Union[Path, str]] = None,
        *args,
        **kwargs
    ) -> str:
        """
        Decorator to dump data to a stream and optionally to a file.

        Args:
            data (Dict[Any, Any]): The data to be written to the serialized stream.
            file_path (Optional[Union[Path, str]], optional): The path to the
            file where the serialized data will be written. If provided, the data
            will be written to the specified file. If not provided, the data
            will only be written to the stream. Defaults to None.

        Returns:
            str: The string containing the written serialized data.

        Raises:
            SerialBusDumperError: If any exception occurs while dumping the data.
        """
        try:
            stream = io.StringIO()
            dump_func(data, stream, *args, **kwargs)

            if file_path is not None:
                with open(file_path, "w") as file:
                    file.write(stream.getvalue())

            return stream.getvalue()
        except Exception as e:
            raise SerialBusDumperError(str(e))

    return standard_dumper_func


@stream_dumper
def json_dumper(
    data: Dict[Any, Any], stream: io.StringIO, *args, **kwargs
) -> str:
    """Writes data to a JSON stream and optionally to a file."""
    json.dump(data, stream, *args, **kwargs)


@stream_dumper
def yaml_dumper(
    data: Dict[Any, Any], stream: io.StringIO, *args, **kwargs
) -> str:
    """Writes data to a YAML stream and optionally to a file."""
    yaml.dump(data, stream, *args, **kwargs)


@stream_dumper
def yml_dumper(
    data: Dict[Any, Any], stream: io.StringIO, *args, **kwargs
) -> str:
    """Writes data to a YAML stream and optionally to a file."""
    yaml.dump(data, stream, *args, **kwargs)


@stream_dumper
def ini_dumper(
    data: Dict[Any, Any], stream: io.StringIO, *args, **kwargs
) -> str:
    """Writes data to an INI stream and optionally to a file."""
    config = configparser.ConfigParser()
    config.read_dict(data)
    config.write(stream, *args, **kwargs)


@stream_dumper
def toml_dumper(
    data: Dict[Any, Any], stream: io.StringIO, *args, **kwargs
) -> str:
    """Writes data to a TOML stream and optionally to a file."""
    toml.dump(data, stream, *args, **kwargs)



import importlib
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Type, TypeVar, Union

# This module is where the 'HashableDict' will be used the most.
# Hence, we import it here and in serial_bus.__init__.py we import
# it from here. Other modules, including client code, should always do
# 'from serial_bus import HashableDict'.
from serial_bus.config import get_config
from serial_bus.exceptions import (
    SerialBusImportError,
    SerialBusTypeError,
    UnsupportedFileFormatError,
)

GLOBAL_CONFIGS = get_config()
_module, _, _cls = GLOBAL_CONFIGS.hashable_dict_cls.rpartition(".")
try:
    hashable_dict_module = importlib.import_module(_module)
    HashableDict = getattr(hashable_dict_module, _cls)
except AttributeError:
    raise SerialBusImportError(
        f"Could not import class {_cls} from module {_module}"
    )

# Creating a generic type variable to be used in type hints for the HashableDict
# class. This is necessary because the class actually assigned to HashableDict is
# not known in advance, as it can be dynamically defined by the user.
HashableDictType = TypeVar("HashableDictType", bound=HashableDict)

if TYPE_CHECKING:
    from serial_bus.models import SerialBusBaseModel


def check_support_by_extension(file_path: Path) -> bool:
    """Checks if the file represented by `file_path` is of a
    supported format.

    Here we take a rather naive approach and only check the file
    extension.

    Args:
        file_path (Path): the path to the file to be checked.
    
    Raises:
        SerialBusTypeError: If `file_path` does not represent a file.

    Returns:
        bool: True if the file is of a supported format, False otherwise.
    """
    if not file_path.is_file():
        raise SerialBusTypeError(
            f"{file_path.absolute()} does not exist or is not a file."
        )
    return file_path.suffix.lstrip(".") in GLOBAL_CONFIGS.supported_formats


def load_file_to_dict(file_path: Union[str, Path]) -> dict:
    """Finds the appropriate loader for the file format and uses it to load the file.

    Args:
        file_path (str, Path): the path to the file to be loaded. If it's a
        string, it will be converted to a Path object.

    Raises:
        SerialBusImportError: If the supplied module with loader
        functions can't be imported.
        UnsupportedFileFormatError: If the file format is not supported.

    Returns:
        dict: the loaded file as a dictionary.
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)

    if not check_support_by_extension(file_path):
        raise UnsupportedFileFormatError(
            f"File extension {file_path.suffix.lstrip('.')} not supported. "
            f"Supported formats are {GLOBAL_CONFIGS.supported_formats}"
        )

    try:
        loader_function = getattr(
            GLOBAL_CONFIGS.loaders_module, f"{file_path.suffix.lstrip('.')}_loader"
        )
    except AttributeError:
        raise SerialBusImportError(
            f"Could not find a loader function with name {file_path.suffix.lstrip('.')}_loader"
        )
    return loader_function(file_path)


def convert_src_file_to(
    src_file: Union[str, Path],
    dst_format: str,
    dst_file: Optional[Union[str, Path]] = None,
    *args,
    **kwargs,
) -> None:
    """
    TODO: change this function behavior to OPTIONALLY save the converted data to a file.
    Converts a serialized file to another format. Saves the content of
    converted data to a file and also returns the converted data as a string.


    Args:
        src_file (str, Path): the path to the file to be converted. Can be either
        a string or a Path object.
        dst_format (str): the format to which the file will be converted.
        dst_file (Optional[Path], optional): the path to the output file. If not
        provided, the output will be written to a file with the same name as the input
        file, but with the new extension. Defaults to None.

        additional arguments and keyword arguments will be passed to the dumper
        function.

    Raises:
        UnsupportedFileFormatError: If the src_file format is not supported.
        SerialBusImportError: If either a loader of dumper function
        can't be found for the src_file format or dst_format respectively.
    """
    if isinstance(src_file, str):
        src_file = Path(src_file)

    if isinstance(dst_file, str):
        dst_file = Path(dst_file)

    loaded_dict = load_file_to_dict(src_file)

    try:
        dumper_function = getattr(GLOBAL_CONFIGS.dumpers_module, f"{dst_format}_dumper")
    except AttributeError:
        raise SerialBusImportError(
            f"Could not find a dumper function with name {dst_format}_dumper"
        )

    dst_file = dst_file or src_file.with_suffix(f".{dst_format}")

    return dumper_function(loaded_dict, dst_file, *args, **kwargs)


def convert_flat_dict_to_hashabledict(dict_obj: dict) -> Type[HashableDictType]:
    """
    Converts a flat dictionary to a HashableDict.
    This function SHOULDN'T be used directly, but rather through
    `convert_dict_to_hashabledict` which will handle the
    conversion of nested dictionaries.

    Args:
        dict_obj (dict): the dictionary to be converted.

    Returns:
        Type[HashableDictType]: the converted dictionary as a HashableDict.
    """
    if not dict_obj:
        return HashableDict()

    if not isinstance(dict_obj, HashableDict):
        dict_obj = HashableDict(dict_obj)

    return dict_obj


def convert_dict_to_hashabledict(dict_obj: dict) -> Type[HashableDictType]:
    """Converts a nested dictionary to a HashableDict.

    Args:
        dict_obj (dict): the dictionary to be converted.

    Returns:
        Type[HashableDictType]: the converted dictionary as a HashableDict.
    """
    for k in dict_obj:
        if isinstance(dict_obj[k], dict):
            convert_dict_to_hashabledict(dict_obj[k])
            dict_obj[k] = convert_flat_dict_to_hashabledict(dict_obj[k])

    return convert_flat_dict_to_hashabledict(dict_obj)


def generate_from_dict(
    loaded_dict: dict,
) -> None:
    """
    Instantiate all SerialBus models from a dictionary.

    Args:
        loaded_dict (dict): the dictionary to be used to create the models.

    Raises:
        DataTypeError: If `loaded_dict` is not a dict.

    """
    for key in loaded_dict:
        model_cls = GLOBAL_CONFIGS.directive_to_model_mapping.get(key)
        if model_cls:
            if isinstance(loaded_dict[key], list):
                for dict_element in loaded_dict[key]:
                    generate_from_dict(dict_element)

            model_cls.create_from_loaded_data(loaded_dict[key])


def generate_from_file(file_path: Union[str, Path]) -> None:
    """
    A very straightforward function to instantiate all models from a file.

    This will only work if there's no need for any kind of pre-processing of the
    data before creating the models.

    If there's a need for pre-processing, then a better approach would be to:
        1 - load the data as a dict by using `load_file_to_dict`
        2 - do the required processing of the loaded data
        3 - create the models from the processed data using `create_all_models_from_dict`

    Args:
        file_path (Union[str, Path]): The Path object or a str that can be converted to a 
        Path object to the file containing serialized data to be loaded and parsed into a 
        dictionary. 
    """
    loaded_data = load_file_to_dict(file_path)
    generate_from_dict(loaded_dict=loaded_data)



from serial_bus.datastore import get_global_data_store
from serial_bus.models import (
    SerialBusBaseModel,
    SerialBusRenderableModel,
)
from serial_bus.utils import HashableDict, HashableDictType

GLOBAL_DATA_STORE = get_global_data_store()


__all__ = [
    "SerialBusBaseModel",
    "SerialBusRenderableModel",
    "HashableDict",
    "HashableDictType",
    "GLOBAL_DATA_STORE",
]



import os
import pytest
from serial_bus.utils import convert_src_file_to
from serial_bus.config import get_config
from serial_bus.exceptions import SerialBusDumperError

# Define the supported formats
CONFIG = get_config()
# formats = CONFIG.supported_formats  #
formats = ["toml", "yaml", "yml", "ini"]


# Define a fixture for the source files
@pytest.fixture
def src_file():
    return "serial_bus/tests/mocked_data/customers.json"


# Define a fixture for the destination files
@pytest.fixture(params=formats, ids=lambda x: f"dst_format_{x}")
def dst_file(request):
    return f"serial_bus/tests/mocked_data/_customers.{request.param}"


def test_format_conversion(src_file, dst_file):
    if src_file != dst_file:
        # Convert the source file to the destination format
        dst_format = os.path.splitext(dst_file)[1].lstrip('.')
        if dst_format == 'ini':
            with pytest.raises(SerialBusDumperError):
                convert_src_file_to(src_file=src_file, dst_format=dst_format, dst_file=dst_file)
        else:
            convert_src_file_to(src_file=src_file, dst_format=dst_format, dst_file=dst_file)

            # Compare the created file with the expected file
            with open(dst_file, "r") as f_created:
                expected_file = f"serial_bus/tests/mocked_data/customers.{dst_format}"
                with open(expected_file, "r") as f_expected:
                    assert f_created.read() == f_expected.read()



"""
This module contains functions for loading data from various file formats.
It includes functions for loading data from JSON, YAML, TOML and INI files. 
Each function takes an input that holds a reference to the serialized data,
then it parses the data, and returns a Python dictionary. By default,
SerialBus finds which loader to use automatically based on the file
extension through utils.load_file_to_dict() function. If the file format is
not supported, an UnsupportedFileFormatError is raised.

If your project uses file formats other than the ones supported here, you can
add support for them by having your own python loader module and creating an
environment variable `LOADERS_MODULE` that points to its dotted path.
The module should contain functions with the following naming convention:
`{format}_loader` where `format` is the file extension.

Each function should take an argument that brings the serialized data,
and return a dictionary. For example, if you have a custom file format with
the extension `.xyz`, you should create a function called `xyz_loader` in your
custom loaders module that could take a Path object, a str representing the
file system path to the file, or even a string containing the serialized
data itself. After parsing that data, your function must return a dictionary.
"""

import configparser
import json
from pathlib import Path
from typing import Callable, Dict, Union, TextIO

import toml
import yaml


def stream_loader(load_func: Callable[..., Dict]):
    """
    This decorator wraps a loader function to allow it to handle different types of input sources.

    Args:
        load_func (Callable[..., Dict]): The function that will be used to load and parse the data.
        This function should take a TextIO object as its first argument, which represents the stream of data to be
        loaded.
        It may also accept additional arguments, which will be passed through from the `standard_loader_func`.
        The function should return a dictionary representing the parsed data.
    """
    def standard_loader_func(source: Union[Path, str, TextIO], *args, **kwargs) -> dict:
        """
        This decorator wraps a loader function to allow it to handle different types of input sources.
        The input source can be a file path (as a string or Path object) or a TextIO object.
        If the source is a file path, the decorator opens the file and passes the file stream to the
        loader function. If the source is a TextIO object, it is passed directly to the loader function.

        Args:
            source (Union[Path, str, TextIO]): The source of the data to be loaded. This can be a file path (as a
            string or Path object), in which case the file will be opened and its contents passed to `load_func`.
            It can also be a TextIO object, in which case it will be passed directly to `load_func`.
            *args: Variable length argument list to be passed to `load_func`.
            **kwargs: Arbitrary keyword arguments to be passed to `load_func`.

        Returns:
            Callable: The decorated loader function.
        """

        if isinstance(source, (str, Path)):
            with open(source, "r") as stream:
                dictionary = load_func(stream, *args, **kwargs)
        else:
            dictionary = load_func(source, *args, **kwargs)

        return dictionary

    return standard_loader_func


@stream_loader
def json_loader(source: Union[Path, str, TextIO], *args, **kwargs) -> dict:
    """Parses a json text stream and returns a dictionary with its contents."""
    return json.load(source, *args, **kwargs)


@stream_loader
def yaml_loader(source: Union[Path, str, TextIO], *args, **kwargs) -> dict:
    """Parses a yaml text stream and returns a dictionary with its contents."""

    # yaml.load doesn't take any *args, **kwargs
    return yaml.load(source, Loader=yaml.FullLoader)


@stream_loader
def toml_loader(source: Union[Path, str, TextIO], *args, **kwargs) -> dict:
    """Parses a toml text stream and returns a dictionary with its contents."""
    return toml.load(source, *args, **kwargs)


@stream_loader
def ini_loader(source: Union[Path, str, TextIO], *args, **kwargs) -> dict:
    """Parses an ini text stream and returns a dictionary with its contents."""
    config = configparser.ConfigParser()
    config.read_file(source, *args, **kwargs)
    dictionary = {s: dict(config.items(s)) for s in config.sections()}
    return dictionary


def yml_loader(source: Union[Path, str, TextIO], *args, **kwargs) -> dict:
    """This is just to support .yml files. Essentially, it's the same as yaml_loader"""
    return yaml_loader(source, *args, **kwargs)



from serial_bus.utils import convert_src_file_to


# converts src_file to YAML and saves it to a file named "yml_customers.yaml"
convert_src_file_to(src_file="customers.json", dst_format="yaml", dst_file="yml_customers.yaml")

# converts src_file to TOML and saves it to a file named "customers.toml"
convert_src_file_to(src_file="customers.json", dst_format="toml")

# raises SerialBusDumperError because the INI format doesn't support lists.
convert_src_file_to(src_file="customers.json", dst_format="ini")


import inspect
from collections import defaultdict
from functools import cache
from typing import TYPE_CHECKING, Any, DefaultDict, Dict, Optional, Type, Union

from serial_bus.custom_collections import SerialBusSortedSet
from serial_bus.exceptions import (
    DataStoreDirectAssignmentError,
    ModelInstanceAlreadyExistsError,
    ModelInstanceDoesNotExistError,
    MultipleModelInstancesReturnedError,
)

if TYPE_CHECKING:
    from serial_bus.models import SerialBusBaseModel


class ModelsGlobalStore:
    """
    Implements a global store for all the models in the application.

    Models are stored in a defaultdict, where the keys are the model
    classes names and the values are SerialBusSortedSet objects.

    The SerialBusSortedSet is a custom collection that stores the
    models in a sorted order, based on the `key` attribute of the model.

    An example of how the structure looks like:
    {
        "Model1": SerialBusSortedSet([Model1_instance1, Model1_instance2, ...]),
        "Model2": SerialBusSortedSet([Model2_instance1, Model2_instance2, ...]),
        ...
    }

    Only one instance of this class is created, and it's shared across the
    entire application. This is done to ensure that all the models are stored
    in the same place and that they can be accessed from anywhere in the
    application. See the `get_global_data_store` function for more details.
    """

    _records = defaultdict(SerialBusSortedSet)

    @property
    def records(
        self,
    ) -> DefaultDict[str, SerialBusSortedSet]:
        return self._records

    @records.setter
    def records(self, _):
        raise DataStoreDirectAssignmentError(
            "Cannot directly assign to attribute 'records' of a ModelsGlobalStore object."
        )

    def as_dict(self) -> Dict[str, Any]:
        """
        Returns regular python dict version of the records attribute.
        This makes it more human-friendly, as well easier to serialize the
        records attribute to a file by using any of the dumper functions in
        the dumpers module.
        """

        dump = {}

        for key, value in self.records.items():
            dump[key] = [record.model_dump() for record in value]

        return dump

    def save(self, obj: "SerialBusBaseModel") -> None:
        """
        Saves a model instance to the global store.

        If a key matching the object's class name is not in self.records, it will
        be added to the store and the model instance will be added to the
        SerialBusSortedSet associated with that key.Otherwise, the model
        instance will be added to the existing obj.__class__ key.

        Args:
            obj (SerialBusBaseModel): the SerialBusBaseModel instance to be saved.

        Raises:
            ModelAlreadyExists: if the model instance is already in the store
            and the `_err_on_duplicate` attribute is set to True.
        """
        cls_name = self._get_cls_name(obj)

        if obj in self.records[cls_name] and obj._err_on_duplicate:
            raise ModelInstanceAlreadyExistsError(
                f"{cls_name}: duplicates not allowed. Make sure there's no other "
                f"{cls_name} with fields {obj._key} associated with values {obj.key}, respectively."
            )

        self.records[cls_name].add(obj)

    def _get_cls_name(self, obj: Union[Type["SerialBusBaseModel"], "SerialBusBaseModel"]) -> str:
        """
        Returns the name of the class of the given object.

        Args:
            obj (Union[Type["SerialBusBaseModel"], "SerialBusBaseModel"]): could be 
            either a class object or an instance of a class.

        Returns:
            str: the name of the class of the given object.
        """
        if inspect.isclass(obj):
            return obj.__name__
        
        return obj.__class__.__name__

    def _search(
        self,
        model_class: Type["SerialBusBaseModel"],
        search_params: Optional[Dict[Any, Any]] = None,
    ) -> SerialBusSortedSet:
        """
        Searches the records of a given model class based on the search_params.
        Currently, only one k:v pair is supported in the search_params.

        Args:
            model_class (Type["SerialBusBaseModel"]): the class of the model to
            be searched. 
            search_params (Optional[Dict[Any, Any]], optional): a single k:v pair
            dictionary with the key being the attribute of the model and the value
            being the value to be searched for. If None, returns the entire
            SerialBusSortedSet associated with model_class key. Defaults to None.

        Returns:
            SerialBusSortedSet: the SerialBusSortedSet containing the records
            that match the search_params.
        """
        cls_name = self._get_cls_name(model_class)

        if search_params:
            # we only take the first k,v pair from search_params
            search_k, value = list(search_params.items())[0]
            return SerialBusSortedSet(
                [x for x in self.records[cls_name] if getattr(x, search_k) == value]
            )

        return self.records[cls_name]

    def filter(
        self,
        model_class: Type["SerialBusBaseModel"],
        search_params: Dict[Any, Any],
    ) -> SerialBusSortedSet:
        """
        Avails of self._search() to filter the records of a given model class
        based on the search_params. Currently, only one k:v pair is supported
        in the search_params.

        Args:
            model_class (Type["SerialBusBaseModel"]): the class of the
            model to be filtered.
            search_params (Dict[Any, Any]): the search parameters to be used
            to filter the records.

        Returns:
            SerialBusSortedSet: the SerialBusSortedSet containing
            the records that match the search_params.
        """
        return self._search(model_class, search_params)

    def get(
        self,
        model_class: Type["SerialBusBaseModel"],
        search_params: Dict[Any, Any],
    ) -> "SerialBusBaseModel":
        """
        Avails of self.filter() method to get a single model instance from
        the global store based on the search_params.

        Currently, only one k:v pair is supported in the search_params.

        Args:
            model_class (Type["SerialBusBaseModel"]): the class of the model
            to be filtered.
            search_params (Dict[Any, Any]): the search parameters to be used
            to filter the records.

        Raises:
            ModelDoesNotExist: if a SerialBusBaseModel does not exist in
            the datastore matching the search_params.
            ModelAlreadyExists: if more than one SerialBusBaseModel exists
            in the datastore matching the search_params.

        Returns:
            SerialBusBaseModel: a single SerialBusBaseModel instance.
        """
        search = self.filter(model_class, search_params)

        if not search:
            raise ModelInstanceDoesNotExistError(
                f"A {model_class.__name__} object was not found matching params: {search_params}"
            )

        if len(search) > 1:
            raise MultipleModelInstancesReturnedError("More than one element found")

        return search[0]

    def get_all_by_class(
        self, model_class: Type["SerialBusBaseModel"]
    ) -> SerialBusSortedSet:
        """
        Avails of self._search() method to get all the records of a given
        model class from the global store.

        Args:
            model_class (Type["SerialBusBaseModel"]): the class of the model

        Returns:
            SerialBusSortedSet: the SerialBusSortedSet containing all the
            records of the given model class.
        """
        return self._search(model_class)


GLOBAL_DATA_STORE = None


@cache
def get_global_data_store():
    """
    Returns the shared instance of the ModelsGlobalStore. It's a cached function,
    meaning that it will only be executed once and the result will be stored in memory.
    This is done to ensure that the same instance of the class is returned every time
    the function is called.
    """

    global GLOBAL_DATA_STORE

    if GLOBAL_DATA_STORE is None:
        GLOBAL_DATA_STORE = ModelsGlobalStore()

    return GLOBAL_DATA_STORE



"""
This module contains utility functions for test cases
"""


def sort_lists_in_dict(d):
    for k, v in d.items():
        if isinstance(v, list):
            v.sort()

        elif isinstance(v, dict):
            sort_lists_in_dict(v)


def get_customer_model_records_for_assertion(customer_model_records):
    for customer_dict in customer_model_records:
        customer_dict["flagged_interests"] = list(
            {interest['category'] for interest in customer_dict["flagged_interests"]})
        sort_lists_in_dict(customer_dict)

    return customer_model_records


def get_customer_data_for_assertion(customers_data):
    for c_data in customers_data:
        sort_lists_in_dict(c_data)

    return customers_data



import os
import pytest

# Because one of the tests will need access to CustomerModel class,
# it's imported here. And because it's imported before we ever try to
# create any model instance, the `MODELS_MODULES` env var doesn't need
# to be set.
from serial_bus.tests.models import CustomerModel
from serial_bus import GLOBAL_DATA_STORE as data_store
from serial_bus.utils import generate_from_file, load_file_to_dict
from serial_bus.tests.utils import get_customer_model_records_for_assertion, get_customer_data_for_assertion


### FIXTURES ###
@pytest.fixture
def products_yaml_file():
    return "serial_bus/tests/mocked_data/products.yaml"


@pytest.fixture
def customers_json_file():
    return "serial_bus/tests/mocked_data/customers.json"


@pytest.fixture
def customers_toml_file():
    return "serial_bus/tests/mocked_data/customers.toml"


@pytest.fixture
def customers_yaml_file():
    return "serial_bus/tests/mocked_data/customers.yaml"


@pytest.fixture
def yml_customers_yaml_file():
    return "serial_bus/tests/mocked_data/yml_customers.yaml"


@pytest.fixture
def rendered_customers_txt_file():
    return "serial_bus/tests/mocked_data/rendered_customers.txt"


@pytest.fixture
def products_data(products_yaml_file):
    return load_file_to_dict(products_yaml_file)['products']


@pytest.fixture
def customers_data(customers_json_file):
    return load_file_to_dict(customers_json_file)['customers']


### TESTS ###
def test_generate_models(products_yaml_file, customers_json_file):
    # We know that all of our models module(s) have already been imported,
    # and this is enough to guarantee the 'directive_to_model_mapping' config
    # will be populated with the correct values. If that was not the case,
    # we'd have to set the 'MODELS_MODULES' env var as shown below:
    # os.environ["MODELS_MODULES"] = "serial_bus.tests.models"

    generate_from_file(products_yaml_file)
    generate_from_file(customers_json_file)


def test_ds_records_keys():
    assert len(data_store.records.keys()) == 2, "data_store.records should have exactly two keys"
    assert 'ProductModel' in data_store.records, "data_store.records should contain a key named 'ProductModel'"
    assert 'CustomerModel' in data_store.records, "data_store.records should contain a key named 'CustomerModel'"


def test_ds_records_values(products_data, customers_data):
    assert data_store.as_dict()['ProductModel'] == products_data

    _customer_records_for_assertion = get_customer_model_records_for_assertion(data_store.as_dict()['CustomerModel'])
    _customer_data = get_customer_data_for_assertion(customers_data)
    assert len(_customer_records_for_assertion) == len(_customer_data)

    for _dict in _customer_records_for_assertion:
        assert _dict in _customer_data, f"{_dict} not found in _customer_data"


def test_rendering(rendered_customers_txt_file):
    os.environ["TEMPLATES_DIR"] = "serial_bus/tests/templates"

    rendered_str = ''

    for customer in CustomerModel.get_all():
        rendered_str += f"Rendered template for customer: {customer.name}:\n"
        rendered_str += (customer.get_rendered_str() or "Customer doesn't have send_ads set to True")
        rendered_str += f"\n{'-' * 50}\n"

    with open(rendered_customers_txt_file, "r") as f:
        assert rendered_str == f.read()



from sortedcontainers import SortedSet


class HashableDict(dict):
    """
    A ridiculously simple implementation of a hashable dictionary.

    All fields in a SerialBus model must be hashable so that the model
    instance itself can be hashable as well. Because Python's native dict
    is not hashable, this class might come in handy for SerialBus model
    fields that have to be dicts.

    Optionally, a code using SerialBus may use its own implementation of a
    hashable dictionary through 'HASHABLE_DICT_MODULE' and 'HASHABLE_DICT_CLASS'
    environment variables. See serial_bus.config.SerialBusConfig
    class for more details.

    NOTE: The requirement for a hashable dictionary is due to the fact that each
    model instance will be stored in a set data structure, implemented by the
    SerialBusSortedSet class.
    """

    def __hash__(self):
        return hash((frozenset(self), frozenset(self.values())))


class SerialBusSortedSet(SortedSet):
    """This class implements a custom SortedSet with the 'key' argument
    set to a function that returns a model object's `key` property.

    NOTE: the `key` property returns a tuple containing the values associated
    with attribute names found in the private `_key` field of a model object.
    For more details, look at serial_bus.models.SerialBusBaseModel
    class and sortedcontainers.SortedSet class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(key=lambda model_obj: model_obj.key, *args, **kwargs)



from typing import Tuple
from pydantic import EmailStr, Field
from serial_bus.models import SerialBusBaseModel, SerialBusRenderableModel


class ProductModel(SerialBusBaseModel):
    _key = ("prod_id",)  # fields that should be used to uniquely ID a model instance
    _directive = "products"  # yaml key that indicates data for instances of this model
    _err_on_duplicate = True  # if data leading to duplicates are present, should we error?

    prod_id: str
    name: str
    category: str


class CustomerModel(SerialBusRenderableModel):
    _key = ("email",)
    _directive = "customers"
    _err_on_duplicate = True

    name: str
    age: int = Field(ge=18, le=100)
    send_ads: bool
    email: EmailStr
    flagged_interests: Tuple[ProductModel, ...]

    @classmethod
    def create(cls, dict_args, *args, **kwargs):
        interests = list()

        for category in dict_args["flagged_interests"]:
            interests += [prod for prod in ProductModel.filter({"category": category})]

        # serial_bus will take care of ultimately converting this to a tuple.
        dict_args["flagged_interests"] = interests

        super().create(dict_args, *args, **kwargs)



"""
Custom exceptions for SerialBus are defined here.

The idea is to have subclasses of SerialBusBaseException for each
possible error that can occur in the library. This way, the user can either
catch specific exceptions and handle them accordingly, or even use
SerialBusBaseException as an umbrella to intercept any SerialBus
specific error.
"""


class SerialBusBaseException(Exception):
    """Parent class for all SerialBus exceptions."""


class ModelInitializationError(SerialBusBaseException):
    """Model instantiation is not possible."""


class SerialBusTypeError(SerialBusBaseException):
    """Data received is of the wrong type"""


class RenderableTemplateError(SerialBusBaseException):
    """An error occurred while trying to render a RenderableSerialBusModel"""


class ModelInstanceDoesNotExistError(SerialBusBaseException):
    """The model does not exist in the datastore."""


class ModelInstanceAlreadyExistsError(SerialBusBaseException):
    """Model already exists in the datastore."""

class MultipleModelInstancesReturnedError(SerialBusBaseException):
    """More than one model instance found in the datastore matching search criteria, when only one should."""


class DataStoreDirectAssignmentError(SerialBusBaseException):
    """Cannot directly assign to attribute 'records' of a ModelsGlobalStore object."""


class UnsupportedFileFormatError(SerialBusBaseException):
    """Unsupported file format"""


class SerialBusImportError(SerialBusBaseException):
    """Error importing a module or class"""


class SerialBusDumperError(SerialBusBaseException):
    """Error dumping data"""
